{-------------------------------------------------------------------}
{ DSDLIST.INC                                                       }
{ Copyright (c) 1993, 1994  Julian M. Bucknall. All Rights Reserved }
{-------------------------------------------------------------------}

{-A FirstThat iterator for cloning a double linked list}
function FailsDListCloning(DL : PAbstractContainer; aData : pointer;
                           NDL : pointer) : boolean;
  var
    NewList : PDList absolute NDL;
    NewData : pointer;
    NoDataError : boolean;
  begin
    FailsDListCloning := true;
    NoDataError := true;
    with NewList^ do
      begin
        if OwnsData then
             NoDataError := DupData(aData, NewData)
        else NewData := aData;
        if NoDataError then
          if InsertBefore(SetAfterLast, NewData) then
                FailsDListCloning := false
          else if OwnsData then DisposeData(NewData);
      end;
  end;

{-A FirstThat iterator for cloning a SORTED double linked list}
function FailsDListSortCloning(DL : PAbstractContainer; aData : pointer;
                               NDL : pointer) : boolean;
  var
    NewList : PDList absolute NDL;
    NewData : pointer;
    NoDataError : boolean;
  begin
    FailsDListSortCloning := true;
    NoDataError := true;
    with NewList^ do
      begin
        if OwnsData then
             NoDataError := DupData(aData, NewData)
        else NewData := aData;
        if NoDataError then
          if InsertSorted(NewData) then
               FailsDListSortCloning := false
          else if OwnsData then DisposeData(NewData);
      end;
  end;

{=TDList==============================================================
A standard doubly-linked list

19Sep93 JMB
======================================================================}
constructor TDList.Init(DataOwner : boolean);
  begin
    if not inherited Init(12, DataOwner) then Fail;
    {These allocations always succeed}
    BF := acNewNode(nil);
    AL := acNewNode(nil);
    BF^.FLink := AL;
    BF^.BLink:= BF;
    AL^.FLink := AL;
    AL^.BLink:= BF;
    Count := 0;
    IsSorted := true;
  end;
{--------}
function  TDList.Clone(DataOwner : boolean;
                       NewCompare : TCompareFunc) : PAbstractContainer;
  var
    NewList : PDList;
    Failed  : boolean;
  begin
    NewList := PDList(NewContainer(DataOwner));
    if Assigned(NewList) then
      begin
        NewList^.SetCompare(NewCompare);
        if (not IsEmpty) then
          begin
            if IsSorted then
                 Failed := FirstThat(FailsDListSortCloning, NewList) <> 0
            else Failed := FirstThat(FailsDListCloning, NewList) <> 0;
            if Failed then
              begin
                Dispose(NewList, Done);
                NewList := nil;
                Error(edsCloneError);
              end;
          end;
      end;
    Clone := NewList;
  end;
{--------}
function  TDList.Delete(Cursor : TListCursor) : TListCursor;
  var
    Temp : PNode;
  begin
    {$IFDEF DEBUG}
    Assert((not IsBeforeFirst(Cursor)) and (not IsAfterLast(Cursor)),
           'TDList.Delete');
    {$ENDIF}
    Temp := PNode(Cursor);
    Cursor := Next(Cursor);
    Temp^.BLink^.FLink := PNode(Cursor);
    PNode(Cursor)^.BLink := Temp^.BLink;
    acDisposeNode(Temp);
    Delete := Cursor;
    if IsEmpty then
      IsSorted := true;
  end;
{--------}
procedure TDList.Empty;
  var
    Cursor : TListCursor;
  begin
    Cursor := Next(SetBeforeFirst);
    while not IsAfterLast(Cursor) do
      Cursor := Erase(Cursor);
    if InDone then
      begin
        acDisposeNode(BF);
        acDisposeNode(AL);
      end;
  end;
{--------}
function  TDList.Erase(Cursor : TListCursor) : TListCursor;
  begin
    if OwnsData then
      DisposeData(Examine(Cursor));
    Erase := Delete(Cursor);
  end;
{--------}
function  TDList.Examine(Cursor : TListCursor) : pointer;
  begin
    {$IFDEF DEBUG}
    Assert((not IsBeforeFirst(Cursor)) and (not IsAfterLast(Cursor)),
           'TDList.Examine');
    {$ENDIF}
    Examine := PNode(Cursor)^.Data;
  end;
{--------}
function  TDList.FirstThat(Action : TIteratorFunc; ExtraData : pointer) : TListCursor;
  var
    Cursor : TListCursor;
    FoundIt : boolean;
  begin
    FirstThat := 0;
    FoundIt := false;
    Cursor := Next(SetBeforeFirst);
    while (not IsAfterLast(Cursor)) and (not FoundIt) do
      if Action(@Self, Examine(Cursor), ExtraData) then
        begin
          FoundIt := true;
          FirstThat := Cursor;
        end
      else
        Cursor := Next(Cursor);
  end;
{--------}
procedure TDList.ForEach(Action : TIteratorProc; ExtraData : pointer);
  var
    Cursor : TListCursor;
  begin
    Cursor := Next(SetBeforeFirst);
    while not IsAfterLast(Cursor) do
      begin
        Action(@Self, Examine(Cursor), ExtraData);
        Cursor := Next(Cursor);
      end;
  end;
{--------}
function TDList.InsertAfter(Cursor : TListCursor; aData : pointer) : boolean;
  var
    Node : PNode;
  begin
    InsertAfter := false;
    {$IFDEF DEBUG}
    Assert(not IsAfterLast(Cursor), 'TDList.InsertAfter');
    {$ENDIF}
    Node := acNewNode(aData);
    if Assigned(Node) then
      begin
        Node^.FLink := PNode(Cursor)^.FLink;
        Node^.BLink:= PNode(Cursor);
        PNode(Cursor)^.FLink := Node;
        Node^.FLink^.BLink := Node;
        IsSorted := false;
        InsertAfter := true;
      end;
  end;
{--------}
function TDList.InsertBefore(Cursor : TListCursor; aData : pointer) : boolean;
  var
    Node : PNode;
  begin
    InsertBefore := false;
    {$IFDEF DEBUG}
    Assert(not IsBeforeFirst(Cursor), 'TDList.InsertBefore');
    {$ENDIF}
    Node := acNewNode(aData);
    if Assigned(Node) then
      begin
        Node^.FLink := PNode(Cursor);
        Node^.BLink:= PNode(Cursor)^.BLink;
        PNode(Cursor)^.BLink := Node;
        Node^.BLink^.FLink := Node;
        IsSorted := false;
        InsertBefore := true;
      end;
  end;
{--------}
function TDList.InsertSorted(aData : pointer) : boolean;
  var
    Walker    : TListCursor;
    CompResult: integer;
  begin
    if not IsSorted then
      begin
        Walker := SetAfterLast;
        InsertSorted := InsertBefore(Walker, aData);
      end
    else {the list is sorted}
      begin
        if Search(Walker, aData) then
          begin
            Error(edsInsertDup);
            InsertSorted := false;
          end
        else
          begin
            InsertSorted := InsertBefore(Walker, aData);
            IsSorted := true;
          end;
      end;
  end;
{--------}
function  TDList.IsAfterLast(Cursor : TListCursor) : boolean;
  begin
    IsAfterLast := (PNode(Cursor) = AL);
  end;
{--------}
function  TDList.IsBeforeFirst(Cursor : TListCursor) : boolean;
  begin
    IsBeforeFirst := (PNode(Cursor) = BF);
  end;
{--------}
procedure TDList.Join(Cursor : TListCursor; List : PDList);
  var
    JoinCursor : TListCursor;
  begin
    {$IFDEF DEBUG}
    Assert(not IsAfterLast(Cursor), 'TDList.Join');
    {$ENDIF}
    if Assigned(List) then
      begin
        if not List^.IsEmpty then
          begin
            List^.AL^.BLink^.FLink := PNode(Cursor)^.FLink;
            PNode(Cursor)^.FLink^.BLink := List^.AL^.BLink;
            PNode(Cursor)^.FLink := List^.BF^.FLink;
            PNode(Cursor)^.FLink^.BLink := PNode(Cursor);
            inc(Count, List^.Count);
            IsSorted := false;
            {patch up List to be empty}
            with List^ do
              begin
                BF^.FLink := AL;
                AL^.BLink := BF;
                Count := 0;
              end;
          end;
        Dispose(List, Done);
      end;
  end;
{--------}
function  TDList.NewContainer(DataOwner : boolean) : PAbstractContainer;
  var
    NC : PAbstractContainer;
  begin
    NC := New(PDList, Init(DataOwner));
    if Assigned(NC) then
      begin
        NC^.SetCompare(Compare);
        NC^.SetDisposeData(DisposeData);
        NC^.SetDupData(DupData);
      end;
    NewContainer := NC;
  end;
{--------}
function  TDList.Next(Cursor : TListCursor) : TListCursor;
  begin
    {$IFDEF DEBUG}
    Assert(not IsAfterLast(Cursor), 'TDList.Next');
    {$ENDIF}
    Next := TListCursor(PNode(Cursor)^.FLink);
  end;
{--------}
function  TDList.Prev(Cursor : TListCursor) : TListCursor;
  begin
    {$IFDEF DEBUG}
    Assert(not IsBeforeFirst(Cursor), 'TDList.Prev');
    {$ENDIF}
    Prev := TListCursor(PNode(Cursor)^.BLink);
  end;
{--------}
function  TDList.Replace(Cursor : TListCursor; aData : pointer) : pointer;
  begin
    {$IFDEF DEBUG}
    Assert((not IsBeforeFirst(Cursor)) and (not IsAfterLast(Cursor)),
           'TDList.Replace');
    {$ENDIF}
    if IsSorted then
      begin
        Replace := Examine(Cursor);
        Cursor := Delete(Cursor);
        if not InsertSorted(aData) then
          {do nothing - it cannot happen};
      end
    else
      with PNode(Cursor)^ do
        begin
          Replace := Data;
          Data := aData;
        end;
  end;
{--------}
function TDList.Search(var Cursor : TListCursor; aData : pointer) : boolean;
  var
    Walker       : TListCursor;
    CompResult   : integer;
    StillLooking : boolean;
    Found        : boolean;
  begin
    Walker := SetBeforeFirst;
    if IsSorted then
      begin
        CompResult := 1;
        while (CompResult > 0) do
          begin
            Walker := Next(Walker);
            if IsAfterLast(Walker) then
                 CompResult := -1
            else CompResult := Compare(aData, Examine(Walker));
          end;
        Cursor := Walker;
        Search := (CompResult = 0);
      end
    else {the list is not sorted}
      begin
        StillLooking := true;
        Found := false;
        while StillLooking and (not Found) do
          begin
            Walker := Next(Walker);
            if IsAfterLast(Walker) then
                 StillLooking := false
            else Found := (Compare(aData, Examine(Walker)) = 0);
          end;
        Cursor := Walker;
        Search := Found;
      end;
  end;
{--------}
function  TDList.SetBeforeFirst : TListCursor;
  begin
    SetBeforeFirst := TListCursor(BF);
  end;
{--------}
function  TDList.SetAfterLast : TListCursor;
  begin
    SetAfterLast := TListCursor(AL);
  end;
{--------}
function  TDList.Split(Cursor : TListCursor) : PDList;
  var
    {$IFDEF MaxIs64KItems}
    TempCount : word;
    {$ELSE}
    TempCount : longint;
    {$ENDIF}
    NewList   : PDList;
    Walker    : TListCursor;
    LastNodeLeftBehind,
    JoinNode,
    LastNode  : PNode;
  begin
    {$IFDEF DEBUG}
    Assert((not IsBeforeFirst(Cursor)) and (not IsAfterLast(Cursor)),
           'TDList.Split');
    {$ENDIF}
    NewList := PDList(NewContainer(OwnsData));
    if not Assigned(NewList) then
      Error(ConstructorError)
    else
      begin
        LastNodeLeftBehind := PNode(Cursor)^.BLink;

        TempCount := 0;
        Walker := Cursor;
        JoinNode := PNode(Walker);
        while not IsAfterLast(Walker) do
          begin
            inc(TempCount);
            Walker := Next(Walker);
          end;

        LastNode := PNode(Prev(Walker));

        JoinNode^.BLink := NewList^.BF;
        NewList^.BF^.FLink := JoinNode;
        LastNode^.FLink := NewList^.AL;
        NewList^.AL^.BLink := LastNode;
        NewList^.Count := TempCount;
        NewList^.IsSorted := IsSorted;

        dec(Count, TempCount);
        LastNodeLeftBehind^.FLink := AL;
        AL^.BLink := LastNodeLeftBehind;
        if IsEmpty then
          IsSorted := true;
      end;
    Split := NewList;
  end;
{---------------------------------------------------------------------}
