{-------------------------------------------------------------------}
{ DSSTACK.INC                                                       }
{ Copyright (c) 1993, 1994  Julian M. Bucknall. All Rights Reserved }
{-------------------------------------------------------------------}

{=TStack==============================================================
A standard stack object.

19Sep93 JMB
======================================================================}
constructor TStack.Init(DataOwner : boolean);
  begin
    if not inherited Init(8, DataOwner) then Fail;
    {This allocation always succeeds}
    Head := acNewNode(nil);
    Head^.Link := Head;
    Count := 0;
  end;
{--------}
function  TStack.Clone(DataOwner : boolean;
                       NewCompare : TCompareFunc) : PAbstractContainer;
  var
    JoinNode,
    NewNode,
    Node     : PNode;
    NewStack : PStack;
    NewData  : pointer;
    DataError: boolean;
  begin
    NewStack := PStack(NewContainer(DataOwner));
    if not Assigned(NewStack) then
      Error(ConstructorError)
    else
      begin
        NewStack^.SetCompare(NewCompare);
        DataError := false;
        JoinNode := NewStack^.Head;
        Node := Head^.Link;
        while (Node <> Head) and (not DataError) do
          begin
            if not DataOwner then
              NewData := Node^.Data
            else
              DataError := not DupData(Node^.Data, NewData);
            if not DataError then
              begin
                NewNode := NewStack^.acNewNode(NewData);
                if not Assigned(NewNode) then
                  begin
                    {note: the error will have been reported already}
                    if DataOwner then
                      DisposeData(NewData);
                    Dispose(NewStack, Done);
                    Clone := nil;
                    Exit;
                  end;
                NewNode^.Link := JoinNode^.Link;
                JoinNode^.Link := NewNode;
                JoinNode := NewNode;
                Node := Node^.Link;
              end;
          end;
        if DataError then
          begin
            Dispose(NewStack, Done);
            NewStack := nil;
            Error(edsCloneError);
          end;
      end;
    Clone := NewStack;
  end;
{--------}
procedure TStack.Empty;
  begin
    if OwnsData then
      while not IsEmpty do DisposeData(Pop)
    else
      while not IsEmpty do Pop;
    if InDone then
      acDisposeNode(Head);
  end;
{--------}
function  TStack.Examine : pointer;
  begin
    {$IFDEF DEBUG}
    Assert(not IsEmpty, 'TStack.Examine');
    {$ENDIF}
    Examine := Head^.Link^.Data;
  end;
{--------}
function  TStack.NewContainer(DataOwner : boolean) : PAbstractContainer;
  var
    NC : PAbstractContainer;
  begin
    NC := New(PStack, Init(DataOwner));
    if Assigned(NC) then
      begin
        NC^.SetCompare(Compare);
        NC^.SetDisposeData(DisposeData);
        NC^.SetDupData(DupData);
      end;
    NewContainer := NC;
  end;
{--------}
function  TStack.Pop : pointer;
  var
    Node : PNode;
  begin
    {$IFDEF DEBUG}
    Assert(not IsEmpty, 'TStack.Pop');
    {$ENDIF}
    Node := Head^.Link;
    Head^.Link := Node^.Link;
    Pop := Node^.Data;
    acDisposeNode(Node);
  end;
{--------}
function  TStack.Push(aData : pointer) : boolean;
  var
    Node : PNode;
  begin
    Push := false;
    Node := acNewNode(aData);
    if Assigned(Node) then
      begin
        Node^.Link := Head^.Link;
        Head^.Link := Node;
        Push := true;
      end;
  end;
{---------------------------------------------------------------------}
