unit EZIntQue;
  {-Example unit defining a priority queue (min/max heap) for integers}

  {Note: the raison d'etre of this object definition is to show how
         easy it is to define an object based on one of EZSTRUCS
         object classes. And one that you don't have to continually
         remember to typecast the objects when you're adding or
         removing them from the container; all the nastiness is
         hidden!}

interface

uses
  Objects,
  EZStrucs;

type
  {A priority queue for storing longints}
  PIntHeap = ^TIntHeap;
  TIntHeap = object
    Queue : TPriorityQueue;

    constructor Init(Ascending : boolean);
      {-Initialise the queue for Ascending order}
    destructor Done;
      {-Destroy the queue}

    function  Count : longint;
      {-Return the number of longints in the queue}
    function  IsEmpty : boolean;
      {-Return true if the queue is empty}
    function  Pop : longint;
      {-Return the longint at the front of the queue after popping it}
    function  Add(const Value : longint) : boolean;
      {-Return true if the longint was added onto the queue}
  end;

implementation

function  IntCompareUp(Data1, Data2 : pointer) : integer;
  var
    L1 : longint absolute Data1;
    L2 : longint absolute Data2;
  begin
    if (L1 < L2) then      IntCompareUp := -1
    else if (L1 = L2) then IntCompareUp := 0
    else                   IntCompareUp := 1
  end;

function  IntCompareDown(Data1, Data2 : pointer) : integer;
  var
    L1 : longint absolute Data1;
    L2 : longint absolute Data2;
  begin
    if (L1 < L2) then      IntCompareDown := 1
    else if (L1 = L2) then IntCompareDown := 0
    else                   IntCompareDown := -1
  end;


{===TIntMaxHeap implementation=========================================}
constructor TIntHeap.Init(Ascending : boolean);
  begin
    if not Queue.Init then
      Fail;
    Queue.SetDisposeData(EZIntDisposeData);
    if Ascending then
         Queue.SetCompare(IntCompareUp)
    else Queue.SetCompare(IntCompareDown);
  end;
{--------}
destructor TIntHeap.Done;
  begin
    Queue.Done;
  end;
{--------}
function TIntHeap.Count : longint;
  begin
    Count := Queue.Count;
  end;
{--------}
function TIntHeap.IsEmpty : boolean;
  begin
    IsEmpty := Queue.IsEmpty;
  end;
{--------}
function TIntHeap.Add(const Value : longint) : boolean;
  begin
    Add := Queue.Append(pointer(Value));
  end;
{--------}
function TIntHeap.Pop : longint;
  begin
    if IsEmpty then
         Pop := 0 {as good a value as any other!}
    else Pop := longint(Queue.Pop);
  end;
{====================================================================}

end.